/*
 REACTIVE AUTONOMOUS BLACKHOLE LIST SERVER
 COPYRIGHT (C) 2005 DEEP LOGIC, INC.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifdef HAVE_CONFIG_H
#include <auto-config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>

#ifdef TIME_WITH_SYS_TIME
#   include <sys/time.h>
#   include <time.h>
#else
#   ifdef HAVE_SYS_TIME_H
#       include <sys/time.h>
#   else
#       include <time.h>
#   endif
#endif

#include <sys/types.h>
#include <sys/resource.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <signal.h>


#include "rabl_client.h"
#include "error.h"
#include "hash.h"
#include "md5.h"
#include "read_config.h"

int
main (int argc, char *argv[])
{
  int i, func = 0;
  char *address = NULL;

  config = read_config(NULL);
  if (!config) {
    fprintf(stderr, "unable to read configuration\n");
    exit(EXIT_FAILURE);
  }

  sockfd = 0;
  for(i=0;i<argc;i++) {
    if (!strcmp(argv[i], "-c") && i+1<argc) {
      func = FUNC_QUERY;
      address = argv[i+1];
    }  

    else if (!strcmp(argv[i], "-s") && i+1<argc) {
      func = FUNC_POST_SPAM;
      address = argv[i+1];
    }

    else if (!strcmp(argv[i], "-d") && i+1<argc) {
      func = FUNC_WATCH_DIR;
      address = argv[i+1];
    }
  }

  if (func) {
    process_function(func, address);
  } else {
    fprintf(stderr, "unrecognized command\n");
    fprintf(stderr, SYNTAX, VERSION);
    return EXIT_FAILURE;
  }

  return EXIT_SUCCESS;
}

int process_function(int func, char *address) {

  if (sockfd == 0) 
    sockfd = establish_connection();
  if (sockfd<=0) 
    return EFAILURE;

  if (func == FUNC_QUERY) {
    if (strchr(address, ':')) 
      return query_ipv6_address(sockfd, address);
    else 
      return query_ipv4_address(sockfd, address);

  } else if (func == FUNC_POST_SPAM) {
    if (strchr(address, ':'))
      return post_ipv6_address(sockfd, address);
    else
      return post_ipv4_address(sockfd, address); 
  } else if (func == FUNC_WATCH_DIR) {
    return watch_directory(address);
  }

  close(sockfd);
  return 0;
}

int watch_directory(const char *dir) {
  DIR *d;
  struct dirent *entry;
  struct hash *ht;
  struct ht_node *node_ht;
  struct ht_c c_ht;
  struct rabl_client_entry *ce;
  time_t last_cleanup = time(NULL);
  char link[1024];
  int cleanup_interval = 600;
  int cache_life = 3600;
  int watch_interval = 1;

  if (_ds_read_attribute(config, "CleanupInterval"))
    cleanup_interval = atoi(_ds_read_attribute(config, "CleanupInterval")); 

  if (_ds_read_attribute(config, "CacheLife"))
    cache_life = atoi(_ds_read_attribute(config, "CacheLife"));

  if (_ds_read_attribute(config, "WatchInterval"))
    watch_interval = atoi(_ds_read_attribute(config, "WatchInterval"));

  signal(SIGPIPE, SIG_IGN);
  ht = ht_create(24593ul);
  if (ht == NULL)
    return EFAILURE;

  while(1) {
    time_t local = time(NULL);
    d = opendir(dir);
    if (d == NULL) {
      fprintf(stderr, "unable to open directory '%s' for reading: %s\n", dir, strerror(errno));
      return EFAILURE;
    }

   node_ht = c_ht_first(ht, &c_ht);
      while(node_ht != NULL) {
        node_ht = c_ht_next(ht, &c_ht);
      }

    while ((entry = readdir (d)) != NULL) {
      if (entry->d_name[0] == '.')
        continue;
      if (!ht_find(ht, entry->d_name)) {
        printf("reporting %s\n", entry->d_name);

        ce = calloc(1, sizeof(struct rabl_client_entry));
        if (ce == NULL) {
          ht_destroy(ht);
          return EFAILURE;
        }
        ce->last_hit = local;

        ht_set(ht, entry->d_name, ce);
        if (process_function(FUNC_POST_SPAM, entry->d_name)) {
          close(sockfd);
          sockfd = 0;
          if (process_function(FUNC_POST_SPAM, entry->d_name)) 
            fprintf(stderr, "unable to post spam for %s: repeated failures\n", entry->d_name);
        }
      }
    }
    closedir(d);
    
    if (local-last_cleanup > cleanup_interval) {
      node_ht = c_ht_first(ht, &c_ht);
      while(node_ht != NULL) {
        ce = node_ht->ptr;
        if (ce == NULL || local-ce->last_hit > cache_life) {
          free(ce);
          snprintf(link, sizeof(link), "%s/%s", dir, node_ht->key);
          printf("removing %s from cache\n", node_ht->key);
          ht_delete(ht, node_ht->key);
          unlink(link);
        }

        node_ht = c_ht_next(ht, &c_ht);
      }
      last_cleanup = local;
    }

    sleep(watch_interval);
  }

  return 0;
}

int establish_connection(void) {
  struct sockaddr_in addr;
  int yes = 1, sockfd;
  char version;
  int port = 1039;

  if (_ds_read_attribute(config, "ServerPort"))
    port = atoi(_ds_read_attribute(config, "ServerPort"));

  sockfd = socket(AF_INET, SOCK_STREAM, 0);

  memset(&addr, 0, sizeof(struct sockaddr_in));
  addr.sin_family = AF_INET;

  addr.sin_addr.s_addr = inet_addr(_ds_read_attribute(config, "ServerAddress"));
  addr.sin_port = htons(port);

  if(connect(sockfd, (struct sockaddr *)&addr, sizeof(struct sockaddr_in))<0) {
    fprintf(stderr, "couldn't connect to %s:%d: %s\n", _ds_read_attribute(config, "ServerAddress"), port, strerror(errno));
    return EFAILURE;
  }
                                                                                                                               
  setsockopt(sockfd,SOL_SOCKET,TCP_NODELAY,&yes,sizeof(int));

  if (!recv(sockfd, &version, 1, 0)) { 
    close(sockfd);
    fprintf(stderr, "connection reset by peer\n");
    return EFAILURE;
  }

  if (version != PROTOCOL_VERSION[0]) {
    fprintf(stderr, "client/server protocol version mismatch.  please upgrade your client\n");
    close(sockfd);
    return EFAILURE;
  }

  authenticated = 0;
  return sockfd;
}

int query_ipv6_address(int sockfd, char *address) {
  char response[1];
  char pt = PT_IPV4;
  struct rabl_data_ipv6 packet;

  memset(&packet, 0, sizeof(struct rabl_data_ipv6));
  if (strlen(address)>(sizeof(packet.address)-1)) { 
    fprintf(stderr, "query address too long\n");
    return EINVAL;
  }

  strcpy(packet.address, address);
  packet.type = DATA_QUERY;
                                                                                                                               
  if (send(sockfd, &pt, sizeof(char), 0)<0)
    return EFAILURE;

  if (send(sockfd, &packet, sizeof(struct rabl_data_ipv4), 0)<0)
    return EFAILURE;

  if (!recv(sockfd, response, 1, 0)) {
    close(sockfd);
    fprintf(stderr, "connection reset by peer\n");
    return EFAILURE;
  }
                                                                                                                               
  printf("%c\n", response[0]);
  return 0;
}

int query_ipv4_address(int sockfd, char *address) {
  char ip[4], response[1];
  char *octet;
  int i;
  char pt = PT_IPV4;
  struct rabl_data_ipv4 packet;
  char *a = strdup(address);

  memset(&packet, 0, sizeof(struct rabl_data_ipv4));
  octet = strtok(a, ".");
  for(i=0;i<4;i++) {
    if (octet != NULL)
      ip[i] = atoi(octet);
    octet = strtok(NULL, ".");
  }
  free(a);

  memcpy(&packet.address, ip, sizeof(u_int32_t));
  packet.type = DATA_QUERY;

  if (send(sockfd, &pt, sizeof(char), 0)<0)
    return EFAILURE;
  if (send(sockfd, &packet, sizeof(struct rabl_data_ipv4), 0)<0)
    return EFAILURE;
  if (!recv(sockfd, response, 1, 0)) {
    close(sockfd);
    fprintf(stderr, "connection reset by peer\n");
    return EFAILURE;
  }

  printf("%c\n", response[0]);
  return 0;
}

int post_ipv6_address(int sockfd, char *address) {
  char response[1];
  char pt = PT_IPV6;
  struct rabl_data_ipv6 packet;

  memset(&packet, 0, sizeof(struct rabl_data_ipv6));
  if (strlen(address)+1>sizeof(packet.address)) 
    return EFAILURE;
  strcpy(packet.address, address);
  packet.type = DATA_POST_SPAM;

  if (!authenticated) {
    if (authenticate(sockfd))
      return EFAILURE;
  }

  if (send(sockfd, &pt, sizeof(char), 0)<0)
    return EFAILURE;
  if (send(sockfd, &packet, sizeof(struct rabl_data_ipv6), 0)<0)
    return EFAILURE;
  if (!recv(sockfd, response, 1, 0)) {
    close(sockfd);
    fprintf(stderr, "connection reset by peer\n");
    return EFAILURE;
  }

  printf("%c\n", response[0]);
  return 0;
}

int post_ipv4_address(int sockfd, char *address) {
  char ip[4], response[1];
  char *octet;
  int i;
  char pt = PT_IPV4;
  struct rabl_data_ipv4 packet;
  char *a = strdup(address);

  octet = strtok(a, ".");
  for(i=0;i<4;i++) {
    if (octet != NULL)
      ip[i] = atoi(octet);
    octet = strtok(NULL, ".");
  }

  free(a);
  memset(&packet, 0, sizeof(struct rabl_data_ipv4));
  memcpy(&packet.address, ip, sizeof(u_int32_t));
  packet.type = DATA_POST_SPAM;

  if (!authenticated) {
    if (authenticate(sockfd))
      return EFAILURE;
  }

  if (send(sockfd, &pt, sizeof(char), 0)<0)
    return EFAILURE;
  if (send(sockfd, &packet, sizeof(struct rabl_data_ipv4), 0)<0)
    return EFAILURE;
  if (!recv(sockfd, response, 1, 0)) {
    close(sockfd);
    fprintf(stderr, "connection reset by peer\n");
    return EFAILURE;
  }

  printf("%c\n", response[0]);

  return 0;
}

int authenticate(int sockfd) {
  struct rabl_data_auth packet;
  char pt = PT_AUTH;
  int recv_len = 0, i;
  char data[16], md5[17];
  MD5_CTX MTX;

  if (_ds_read_attribute(config, "ServerUID") == NULL ||
      _ds_read_attribute(config, "ServerSecret") == NULL) {
    fprintf(stderr, "no authentication information set in rabl_client.conf\n");
    exit(EXIT_FAILURE);
  }
    
  packet.uid = atoi(_ds_read_attribute(config, "ServerUID"));
  if (send(sockfd, &pt, sizeof(char), 0)<0)
    return EFAILURE;
  if (send(sockfd, &packet, sizeof(struct rabl_data_auth), 0)<0)
    return EFAILURE;

  while(recv_len<16) {
    i = recv(sockfd, data+recv_len, 16-recv_len, 0);
    if (!i) {
      close(sockfd);
      fprintf(stderr, "connection reset by peer\n");
      return EFAILURE;
    }    
    if (i>0)
      recv_len += i;
  }

  MD5Init (&MTX);
  MD5Update (&MTX, data, 16);
  MD5Update (&MTX, _ds_read_attribute(config, "ServerSecret"), strlen(_ds_read_attribute(config, "ServerSecret")));
  MD5Final (md5, &MTX);

  if (send(sockfd, md5, 16, 0)<0)
    return EFAILURE;

  if (! recv(sockfd, data, 1, 0)) {
    close(sockfd);
    fprintf(stderr, "connection reset by peer\n");
    return EFAILURE;
  }

  if (data[0] == 'A') {
    authenticated = 1; 
    return 0;
  }

  fprintf(stderr, "authentication failed with error '%c'\n", data[0]);
  close(sockfd);
  return EFAILURE;
}

